__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""Geofabrik Data Config

This module contains functions to set config for Geofabrik data.

"""


import json
import logging
import os
from collections import namedtuple

import geopandas as gpd
import pandas as pd

from earth_osm.gfk_download import download_sitemap

logger = logging.getLogger("osm_data_extractor")
logger.setLevel(logging.DEBUG)


pkg_data_dir = os.path.join(os.path.dirname(__file__), "data")
sitemap = download_sitemap(False, pkg_data_dir)

with open(sitemap, encoding='utf8') as f:
    d = json.load(f)

row_list = []
for feature in d['features']:
    row_list.append(feature['properties'])
df = pd.DataFrame(row_list)
# Add short code column
col1 = df['iso3166-1:alpha2'].apply(lambda x: '-'.join(x) if type(x) == list else x)
col2 = df['iso3166-2'].apply(lambda x: '-'.join(x) if type(x) == list else x)
df['short_code'] = col1.combine_first(col2)

def get_geom_sitemap():
    geom_sitemap = download_sitemap(True, pkg_data_dir)
    return gpd.read_file(geom_sitemap)


def get_root_list():
    """
    Returns a list of regions without parents (i.e continents)
    """
    return list(df.loc[df['parent'].isna(), 'id'])

def get_all_valid_list():
    """
    Returns a list of all valid region ids
    """
    return list(df.loc[~df['short_code'].isna(), 'short_code']) + list(df['id'])

def get_all_regions_dict(level=0):
    """
    It takes a level argument, and returns a dictionary of all regions, grouped by their parent region
    
    Args:
        level: 0 = all regions, 1 = world regions, 2 = local regions, defaults to 0
        A dictionary of dictionaries.
    """
    by_parent = df.groupby("parent", as_index=False)[["id", "short_code"]]
    parent_dict = by_parent.groups
    root = get_root_list()
    world_dict = {}
    local_dict = {}
    def dict_by_key(key): return by_parent.get_group(key).set_index('id').T.to_dict('records')[0]
    for key in parent_dict:
        if key in root:
            world_dict[key] = dict_by_key(key)
        else:
            local_dict[key] = dict_by_key(key)
    
    if level == 0:
        return {**world_dict, **local_dict}
    if level == 1:
        return world_dict
    if level == 2:
        return local_dict


def view_regions(level=0):                                                                              
    """
    Takes the `all_regions` dictionary and returns a new dictionary with the same keys, but with
    the values being the `region_id`s of the regions
    """
    all_dict = get_all_regions_dict(level)
    view_df = pd.DataFrame.from_dict(
        {
            (i, j): all_dict[i][j]
                            for i in all_dict.keys() 
            for j in all_dict[i].keys()
        },
        orient='index',
    )
    view_df.index = pd.MultiIndex.from_tuples(view_df.index, names=['parent', 'id'])
    print(view_df.to_string())


def get_region_dict(id):
    """
    Takes a region id (eg. germany) and returns a ditctionary consisting of
    strings 'id', 'name', 'parent', 'short_code' and dictionary of 'urls'
    Raises error if id is not found
    """
    return (
        df.loc[df['id'] == id]
        .drop('iso3166-1:alpha2', axis=1)
        .drop('iso3166-2', axis=1)
        .to_dict('records')[0]
    )


def get_id_by_code(code):
    """
    Takes a region code (eg. DE) and returns its id (eg. germany)
    Supresses error if id is not found
    """
    try:
        return df.loc[df['short_code']== code, 'id'].item()
    except (KeyError, ValueError):
        logger.debug(f'{code} not found')
        return None


def get_code_by_id(id):
    """
    Takes a region id (eg. germany) and returns its code (eg. DE)
    Supresses error if id is not found
    """
    try:
        c_dict = get_region_dict(id)
    except (KeyError, IndexError):
        logger.debug(f'{id} not found')
        return None

    code = str(c_dict["short_code"])
    return code


def get_id_by_str(region_str):
    """
    Takes a region id or code (eg. DE, germany) and returns its id (eg. germany)
    Raises error if the string is not a valid id or code
    """
    # if region_str is region code
    id = get_id_by_code(region_str) 
    if id is not None: 
        return id
    else:
    # if region_str is region id
        code = get_code_by_id(region_str)
        if code is not None:
            return region_str
        else:
            logger.error(f'{region_str} not found. check view_region()')
            raise KeyError(f'{region_str} not found. check view_region()')


def get_region_tuple(region_str):
    """
    Takes a region id or code (eg. DE, germany) and returns a named tuple with 
    'id', 'name', 'short', 'parent', 'short_code' and dictionary of 'urls'
    The 'short' field is an iso code if found otherwise the id is used.
    iso3166-1:alpha2 is used for countries, iso3166-2 is used for sub-divisions
    Raises error if the string is not a valid id or code
    """
    id = get_id_by_str(region_str)
    d = get_region_dict(id)
    d['short'] = d.pop('short_code')
    if str(d['short']) == 'nan':
        logger.warning(f'code not found for {id} so using id')
        d['short'] = d['id']

    Region = namedtuple('Region', d)
    region_tuple = Region(**d)
    return region_tuple




__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""Utilities functions for earth_osm

This module contains utilities functions for handling OSM data.

"""


import logging
import os
import pandas as pd
import geopandas as gpd
from shapely.geometry import LineString, Point, Polygon

from earth_osm.config import primary_feature_element

logger = logging.getLogger("osm_data_extractor")
logger.setLevel(logging.INFO)

# geo_crs: EPSG:4326  # general geographic projection, not used for metric measures. "EPSG:4326" is the standard used by OSM and google maps
# distance_crs: EPSG:3857  # projection for distance measurements only. Possible recommended values are "EPSG:3857" (used by OSM and Google Maps)
# area_crs: ESRI:54009  # projection for area measurements only. Possible recommended values are Global Mollweide "ESRI:54009"

def lonlat_lookup(df_way, primary_data):
    """
    Lookup refs and convert to list of longlats
    """
    if "refs" not in df_way.columns:
        logger.warning("refs column not found")

    def look(ref):
        lonlat_row = list(map(lambda r: tuple(primary_data["Node"][str(r)]["lonlat"]), ref))
        return lonlat_row

    lonlat_list = df_way["refs"].apply(look)

    return lonlat_list


def convert_ways_points(df_way, primary_data):
    """
    Convert Ways to Point Coordinates
    """
    lonlat_list = lonlat_lookup(df_way, primary_data)
    way_polygon = list(
        map(
            lambda lonlat: Polygon(lonlat) if len(lonlat) >= 3 else Point(lonlat[0]),
            lonlat_list,
        )
    )
    area_column = list(
        map(
            int,
            round(
                gpd.GeoSeries(way_polygon)
                .set_crs("EPSG:4326")
                .to_crs("ESRI:54009")
                .area,
                -1,
            ),
        )
    )  # TODO: Rounding should be done in cleaning scripts

    def find_center_point(p):
        if p.geom_type == "Polygon":
            center_point = p.centroid
        else:
            center_point = p
        return list((center_point.x, center_point.y))

    lonlat_column = list(map(find_center_point, way_polygon))

    # df_way.drop("refs", axis=1, inplace=True, errors="ignore")
    df_way.insert(0, "Area", area_column)
    df_way.insert(0, "lonlat", lonlat_column)


def convert_ways_lines(df_way, primary_data):
    """
    Convert Ways to Line Coordinates

    Args:

    """
    lonlat_list = lonlat_lookup(df_way, primary_data)
    lonlat_column = lonlat_list
    df_way.insert(0, "lonlat", lonlat_column)

    way_linestring = map(lambda lonlats: LineString(lonlats), lonlat_list)
    length_column = (
        gpd.GeoSeries(way_linestring)
        .set_crs("EPSG:4326")
        .to_crs("EPSG:3857")
        .length
    )

    df_way.insert(0, "Length", length_column)


def convert_pd_to_gdf_nodes(df_way):
    """
    Convert Nodes Pandas Dataframe to GeoPandas Dataframe

    Args:
        df_way: Pandas Dataframe

    Returns:
        GeoPandas Dataframe
    """
    gdf = gpd.GeoDataFrame(
        df_way, geometry=[Point(x, y) for x, y in df_way.lonlat], crs="EPSG:4326"
    )
    gdf.drop(columns=["lonlat"], inplace=True)
    return gdf


def convert_pd_to_gdf_lines(df_way):
    """
    Convert Lines Pandas Dataframe to GeoPandas Dataframe

    Args:
        df_way: Pandas Dataframe

    Returns:
        GeoPandas Dataframe
    """

    gdf = gpd.GeoDataFrame(
        df_way, geometry=[LineString(x) for x in df_way.lonlat], crs="EPSG:4326"
    )
    gdf.drop(columns=["lonlat"], inplace=True)
    return gdf


def write_csv(df_feature, outputfile_partial, feature_name, out_aggregate, fn_name):
    """Create csv file. Optimized for large files as write on disk in chunks"""
    if out_aggregate:
        output_path = os.path.join(outputfile_partial, f"all_{feature_name}s" + ".csv")
        df_feature.to_csv(
            output_path, index=False, header= not os.path.exists(output_path), mode="a",
        )  # Generate CSV
    else:
        output_path = os.path.join(outputfile_partial, f"{fn_name}_{feature_name}s" + ".csv")
        df_feature.to_csv(
            output_path,
        )  # Generate CSV


def write_geojson(gdf_feature, outputfile_partial, feature_name, out_aggregate, fn_name):
    """Create geojson file. Optimized for large files as write on disk in chunks"""
    if out_aggregate:
        output_path = os.path.join(outputfile_partial, f"all_{feature_name}s" + ".geojson")
        gdf_feature.to_file(
            output_path, driver="GeoJSON", index=False, mode="a" if os.path.exists(output_path) else "w"
        )  # Generate GeoJson
    else:
        output_path = os.path.join(outputfile_partial, f"{fn_name}_{feature_name}s" + ".geojson")
        gdf_feature.to_file(
            output_path, driver="GeoJSON"
        )  # Generate GeoJson


def output_creation(df_feature, primary_name, feature_name, region_list, data_dir, out_format, out_aggregate):
    """
    Output CSV and GeoJSON files for each region

    Args:
        df_feature: _description_
        primary_name: _description_
        feature_name: _description_
        region_list: _description_
    """
    def filenamer(cc_list):
        if len(cc_list) == 1:
            return str(cc_list[0].short)
        else:
            # TODO: Fix filenamer
            raise NotImplementedError

    outputfile_partial = os.path.join(data_dir, "out")  # Output file directory
    fn_name = filenamer(region_list) # country code e.g. BJ

    if not os.path.exists(outputfile_partial):
        os.makedirs(
            outputfile_partial, exist_ok=True
        )  # create raw directory

    df_feature.reset_index(drop=True, inplace=True)

    # Generate Files

    if df_feature.empty:
        logger.warning(f"All feature data frame empty for {feature_name}")
        return None

    if "csv" in out_format:
        write_csv(df_feature, outputfile_partial, feature_name, out_aggregate, fn_name)

    if "geojson" in out_format:
        if primary_feature_element[primary_name][feature_name] == "way":
            gdf_feature = convert_pd_to_gdf_lines(df_feature)
        else:
            gdf_feature = convert_pd_to_gdf_nodes(df_feature)

        try:
            gdf_feature.drop(columns=["refs"], inplace=True)
        except:
            pass

        logger.info("Writing GeoJSON file")
        write_geojson(gdf_feature, outputfile_partial, feature_name, out_aggregate, fn_name)




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: earth_osm/osmpbf/osmformat.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n earth_osm/osmpbf/osmformat.proto\x12\x06OSMPBF\"\x87\x02\n\x0bHeaderBlock\x12 \n\x04\x62\x62ox\x18\x01 \x01(\x0b\x32\x12.OSMPBF.HeaderBBox\x12\x19\n\x11required_features\x18\x04 \x03(\t\x12\x19\n\x11optional_features\x18\x05 \x03(\t\x12\x16\n\x0ewritingprogram\x18\x10 \x01(\t\x12\x0e\n\x06source\x18\x11 \x01(\t\x12%\n\x1dosmosis_replication_timestamp\x18  \x01(\x03\x12+\n#osmosis_replication_sequence_number\x18! \x01(\x03\x12$\n\x1cosmosis_replication_base_url\x18\" \x01(\t\"F\n\nHeaderBBox\x12\x0c\n\x04left\x18\x01 \x02(\x12\x12\r\n\x05right\x18\x02 \x02(\x12\x12\x0b\n\x03top\x18\x03 \x02(\x12\x12\x0e\n\x06\x62ottom\x18\x04 \x02(\x12\"\xd2\x01\n\x0ePrimitiveBlock\x12(\n\x0bstringtable\x18\x01 \x02(\x0b\x32\x13.OSMPBF.StringTable\x12.\n\x0eprimitivegroup\x18\x02 \x03(\x0b\x32\x16.OSMPBF.PrimitiveGroup\x12\x18\n\x0bgranularity\x18\x11 \x01(\x05:\x03\x31\x30\x30\x12\x15\n\nlat_offset\x18\x13 \x01(\x03:\x01\x30\x12\x15\n\nlon_offset\x18\x14 \x01(\x03:\x01\x30\x12\x1e\n\x10\x64\x61te_granularity\x18\x12 \x01(\x05:\x04\x31\x30\x30\x30\"\xb7\x01\n\x0ePrimitiveGroup\x12\x1b\n\x05nodes\x18\x01 \x03(\x0b\x32\x0c.OSMPBF.Node\x12!\n\x05\x64\x65nse\x18\x02 \x01(\x0b\x32\x12.OSMPBF.DenseNodes\x12\x19\n\x04ways\x18\x03 \x03(\x0b\x32\x0b.OSMPBF.Way\x12#\n\trelations\x18\x04 \x03(\x0b\x32\x10.OSMPBF.Relation\x12%\n\nchangesets\x18\x05 \x03(\x0b\x32\x11.OSMPBF.ChangeSet\"\x18\n\x0bStringTable\x12\t\n\x01s\x18\x01 \x03(\x0c\"q\n\x04Info\x12\x13\n\x07version\x18\x01 \x01(\x05:\x02-1\x12\x11\n\ttimestamp\x18\x02 \x01(\x03\x12\x11\n\tchangeset\x18\x03 \x01(\x03\x12\x0b\n\x03uid\x18\x04 \x01(\x05\x12\x10\n\x08user_sid\x18\x05 \x01(\r\x12\x0f\n\x07visible\x18\x06 \x01(\x08\"\x8a\x01\n\tDenseInfo\x12\x13\n\x07version\x18\x01 \x03(\x05\x42\x02\x10\x01\x12\x15\n\ttimestamp\x18\x02 \x03(\x12\x42\x02\x10\x01\x12\x15\n\tchangeset\x18\x03 \x03(\x12\x42\x02\x10\x01\x12\x0f\n\x03uid\x18\x04 \x03(\x11\x42\x02\x10\x01\x12\x14\n\x08user_sid\x18\x05 \x03(\x11\x42\x02\x10\x01\x12\x13\n\x07visible\x18\x06 \x03(\x08\x42\x02\x10\x01\"\x17\n\tChangeSet\x12\n\n\x02id\x18\x01 \x02(\x03\"l\n\x04Node\x12\n\n\x02id\x18\x01 \x02(\x12\x12\x10\n\x04keys\x18\x02 \x03(\rB\x02\x10\x01\x12\x10\n\x04vals\x18\x03 \x03(\rB\x02\x10\x01\x12\x1a\n\x04info\x18\x04 \x01(\x0b\x32\x0c.OSMPBF.Info\x12\x0b\n\x03lat\x18\x08 \x02(\x12\x12\x0b\n\x03lon\x18\t \x02(\x12\"{\n\nDenseNodes\x12\x0e\n\x02id\x18\x01 \x03(\x12\x42\x02\x10\x01\x12$\n\tdenseinfo\x18\x05 \x01(\x0b\x32\x11.OSMPBF.DenseInfo\x12\x0f\n\x03lat\x18\x08 \x03(\x12\x42\x02\x10\x01\x12\x0f\n\x03lon\x18\t \x03(\x12\x42\x02\x10\x01\x12\x15\n\tkeys_vals\x18\n \x03(\x05\x42\x02\x10\x01\"\x85\x01\n\x03Way\x12\n\n\x02id\x18\x01 \x02(\x03\x12\x10\n\x04keys\x18\x02 \x03(\rB\x02\x10\x01\x12\x10\n\x04vals\x18\x03 \x03(\rB\x02\x10\x01\x12\x1a\n\x04info\x18\x04 \x01(\x0b\x32\x0c.OSMPBF.Info\x12\x10\n\x04refs\x18\x08 \x03(\x12\x42\x02\x10\x01\x12\x0f\n\x03lat\x18\t \x03(\x12\x42\x02\x10\x01\x12\x0f\n\x03lon\x18\n \x03(\x12\x42\x02\x10\x01\"\xe0\x01\n\x08Relation\x12\n\n\x02id\x18\x01 \x02(\x03\x12\x10\n\x04keys\x18\x02 \x03(\rB\x02\x10\x01\x12\x10\n\x04vals\x18\x03 \x03(\rB\x02\x10\x01\x12\x1a\n\x04info\x18\x04 \x01(\x0b\x32\x0c.OSMPBF.Info\x12\x15\n\troles_sid\x18\x08 \x03(\x05\x42\x02\x10\x01\x12\x12\n\x06memids\x18\t \x03(\x12\x42\x02\x10\x01\x12.\n\x05types\x18\n \x03(\x0e\x32\x1b.OSMPBF.Relation.MemberTypeB\x02\x10\x01\"-\n\nMemberType\x12\x08\n\x04NODE\x10\x00\x12\x07\n\x03WAY\x10\x01\x12\x0c\n\x08RELATION\x10\x02\x42\x0f\n\rcrosby.binary')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'earth_osm.osmpbf.osmformat_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\rcrosby.binary'
  _DENSEINFO.fields_by_name['version']._options = None
  _DENSEINFO.fields_by_name['version']._serialized_options = b'\020\001'
  _DENSEINFO.fields_by_name['timestamp']._options = None
  _DENSEINFO.fields_by_name['timestamp']._serialized_options = b'\020\001'
  _DENSEINFO.fields_by_name['changeset']._options = None
  _DENSEINFO.fields_by_name['changeset']._serialized_options = b'\020\001'
  _DENSEINFO.fields_by_name['uid']._options = None
  _DENSEINFO.fields_by_name['uid']._serialized_options = b'\020\001'
  _DENSEINFO.fields_by_name['user_sid']._options = None
  _DENSEINFO.fields_by_name['user_sid']._serialized_options = b'\020\001'
  _DENSEINFO.fields_by_name['visible']._options = None
  _DENSEINFO.fields_by_name['visible']._serialized_options = b'\020\001'
  _NODE.fields_by_name['keys']._options = None
  _NODE.fields_by_name['keys']._serialized_options = b'\020\001'
  _NODE.fields_by_name['vals']._options = None
  _NODE.fields_by_name['vals']._serialized_options = b'\020\001'
  _DENSENODES.fields_by_name['id']._options = None
  _DENSENODES.fields_by_name['id']._serialized_options = b'\020\001'
  _DENSENODES.fields_by_name['lat']._options = None
  _DENSENODES.fields_by_name['lat']._serialized_options = b'\020\001'
  _DENSENODES.fields_by_name['lon']._options = None
  _DENSENODES.fields_by_name['lon']._serialized_options = b'\020\001'
  _DENSENODES.fields_by_name['keys_vals']._options = None
  _DENSENODES.fields_by_name['keys_vals']._serialized_options = b'\020\001'
  _WAY.fields_by_name['keys']._options = None
  _WAY.fields_by_name['keys']._serialized_options = b'\020\001'
  _WAY.fields_by_name['vals']._options = None
  _WAY.fields_by_name['vals']._serialized_options = b'\020\001'
  _WAY.fields_by_name['refs']._options = None
  _WAY.fields_by_name['refs']._serialized_options = b'\020\001'
  _WAY.fields_by_name['lat']._options = None
  _WAY.fields_by_name['lat']._serialized_options = b'\020\001'
  _WAY.fields_by_name['lon']._options = None
  _WAY.fields_by_name['lon']._serialized_options = b'\020\001'
  _RELATION.fields_by_name['keys']._options = None
  _RELATION.fields_by_name['keys']._serialized_options = b'\020\001'
  _RELATION.fields_by_name['vals']._options = None
  _RELATION.fields_by_name['vals']._serialized_options = b'\020\001'
  _RELATION.fields_by_name['roles_sid']._options = None
  _RELATION.fields_by_name['roles_sid']._serialized_options = b'\020\001'
  _RELATION.fields_by_name['memids']._options = None
  _RELATION.fields_by_name['memids']._serialized_options = b'\020\001'
  _RELATION.fields_by_name['types']._options = None
  _RELATION.fields_by_name['types']._serialized_options = b'\020\001'
  _HEADERBLOCK._serialized_start=45
  _HEADERBLOCK._serialized_end=308
  _HEADERBBOX._serialized_start=310
  _HEADERBBOX._serialized_end=380
  _PRIMITIVEBLOCK._serialized_start=383
  _PRIMITIVEBLOCK._serialized_end=593
  _PRIMITIVEGROUP._serialized_start=596
  _PRIMITIVEGROUP._serialized_end=779
  _STRINGTABLE._serialized_start=781
  _STRINGTABLE._serialized_end=805
  _INFO._serialized_start=807
  _INFO._serialized_end=920
  _DENSEINFO._serialized_start=923
  _DENSEINFO._serialized_end=1061
  _CHANGESET._serialized_start=1063
  _CHANGESET._serialized_end=1086
  _NODE._serialized_start=1088
  _NODE._serialized_end=1196
  _DENSENODES._serialized_start=1198
  _DENSENODES._serialized_end=1321
  _WAY._serialized_start=1324
  _WAY._serialized_end=1457
  _RELATION._serialized_start=1460
  _RELATION._serialized_end=1684
  _RELATION_MEMBERTYPE._serialized_start=1639
  _RELATION_MEMBERTYPE._serialized_end=1684
# @@protoc_insertion_point(module_scope)



from earth_osm.eo import get_osm_data


def test_base():
    get_osm_data(
        region_list=["benin"],
        primary_name="power",
        feature_list=["tower"],
        update=True,
        mp=True,
        out_format=["csv", "geojson"],
        out_aggregate=False,
    )



__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""Geofabrik Data Download

This module contains functions to download Geofabrik data.

"""


import hashlib
import logging
import os
import shutil

import requests
import urllib3
from tqdm.auto import tqdm

logger = logging.getLogger("osm_geo")
logger.setLevel(logging.INFO)

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def earth_downloader(url, dir):
    """
    Download file from url to dir

    Args:
        url (str): url to download
        dir (str): directory to download to

    Returns:
        str: filepath of downloaded file
    """
    filename = os.path.basename(url)
    filepath = os.path.join(dir, filename)
    logger.info(f"{filename} downloading to {filepath}")
    os.makedirs(os.path.dirname(filepath), exist_ok=True)  #  create download dir
    with requests.get(url, stream=True, verify=False) as r:
        if r.status_code == 200:
            # url properly found, thus execute as expected
            file_size = int(r.headers.get('Content-Length', 0))
            desc = "(Unknown total file size)" if file_size == 0 else ""
            with tqdm.wrapattr(r.raw, "read", total=file_size, desc=desc, leave=False) as raw:
                with open(filepath, "wb") as f:
                    shutil.copyfileobj(raw, f)
        else:
            # error status code: file not found
            logger.error(
                f"Error code: {r.status_code}. File {filename} not downloaded from {url}"
            )
            filepath = None
    return filepath


def download_pbf(url, update, data_dir):

    dir = os.path.join(data_dir, "pbf")
    pbf_filename = os.path.basename(url)
    pbf_filepath = os.path.join(dir, pbf_filename)

    # TODO: multi-part download each file for parallel downloading... (pip install pySmartDL)
    if not os.path.exists(pbf_filepath):
        # download file
        d_filepath = earth_downloader(url, dir)
        assert d_filepath == pbf_filepath
    else:
        logger.debug(f"{pbf_filename} already exists in {pbf_filepath}")

    return pbf_filepath


# TODO: fix update param
def download_sitemap(geom, pkg_data_dir):
    geofabrik_geo = "https://download.geofabrik.de/index-v1.json"
    geofabrik_nogeo = "https://download.geofabrik.de/index-v1-nogeom.json"
    geofabrik_sitemap_url = geofabrik_geo if geom else geofabrik_nogeo
    logger.info('Downloading Sitemap')
    sitemap_file = earth_downloader(geofabrik_sitemap_url, pkg_data_dir)

    return sitemap_file


# def download_pbf(country_code, update, verify):
# if verify is True:
#     if verify_pbf(PBF_inputfile, geofabrik_url, update) is False:
#         logger.warning(f"md5 mismatch, deleting {geofabrik_filename}")
#         if os.path.exists(PBF_inputfile):
#             os.remove(PBF_inputfile)

#         download_pbf(country_code, update=False, verify=False)  # Only try downloading once

# return PBF_inputfile


# verified_pbf = []


# def verify_pbf(PBF_inputfile, geofabrik_url, update):
#     if PBF_inputfile in verified_pbf:
#         return True

#     geofabrik_md5_url = geofabrik_url + ".md5"
#     PBF_md5file = PBF_inputfile + ".md5"

#     def calculate_md5(fname):
#         hash_md5 = hashlib.md5()
#         with open(fname, "rb") as f:
#             for chunk in iter(lambda: f.read(4096), b""):
#                 hash_md5.update(chunk)
#         return hash_md5.hexdigest()

#     if update is True or not os.path.exists(PBF_md5file):
#         with requests.get(geofabrik_md5_url, stream=True, verify=False) as r:
#             with open(PBF_md5file, "wb") as f:
#                 shutil.copyfileobj(r.raw, f)

#     local_md5 = calculate_md5(PBF_inputfile)

#     with open(PBF_md5file) as f:
#         contents = f.read()
#         remote_md5 = contents.split()[0]

#     if local_md5 == remote_md5:
#         verified_pbf.append(PBF_inputfile)
#         return True
#     else:
#         return False



__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""Filter the Extracted OSM Data

This module filters the extracted OSM data.

"""

import json
import logging
import os
from datetime import datetime

from earth_osm.config import primary_feature_element
from earth_osm.extract import filter_pbf
from earth_osm.gfk_download import download_pbf
from earth_osm.osmpbf import Node, Relation, Way

logging.basicConfig()
logger=logging.getLogger(__name__)
#logger.setLevel(logging.INFO)
logger.setLevel(logging.WARNING)


def feature_filter(primary_data, filter_tuple = ('power', 'line')):

    if set(primary_data.keys()) != set(['Node', 'Way', 'Relation']):
        logger.error('malformed primary_data')

    feature_data = {'Node': {}, 'Way': {}, 'Relation': {}}
    for element in list(feature_data.keys()):
        for id in primary_data[element]:
            if filter_tuple in primary_data[element][id]["tags"].items():
                feature_data[element][id] = primary_data[element][id]
    return feature_data


def run_feature_filter(primary_dict, feature_name):
    primary_name = primary_dict['Metadata']['primary_feature']
    filter_tuple = (primary_name, feature_name)
    primary_data = primary_dict['Data']

    feature_data = feature_filter(primary_data, filter_tuple)

    metadata = {
        'filter_date': str(datetime.now().isoformat()),
        'filter_tuple': json.dumps(filter_tuple),
    }

    feature_dict = {
        'Metadata': metadata,
        'Data':feature_data
        }

    return feature_dict

def run_primary_filter(PBF_inputfile, primary_file, primary_name, multiprocess):
    logger.info('New Pre-Filter Data')
    logger.info('Load OSM data from '+ PBF_inputfile+'\n')

    feature_list = list(primary_feature_element[primary_name].keys())
    pre_filter = {
        Node: {primary_name: feature_list},
        Way: {primary_name: feature_list},
        Relation: {primary_name: feature_list},
    }

    primary_data = filter_pbf(PBF_inputfile, pre_filter, multiprocess)

    metadata = {
        'filter_date': str(datetime.now().isoformat()),
        'primary_feature': primary_name,
    }
    primary_dict = {
        'Metadata': metadata,
        'Data': primary_data
    }
    # Save primary_dict
    with open(primary_file, "w", encoding="utf-8") as target:
        json.dump(
            primary_dict,
            target,
            ensure_ascii=False,
            indent=4,
            separators=(",", ":"),
        )

    return primary_dict


def get_filtered_data(region, primary_name, feature_name, mp, update, data_dir):
    geofabrik_pbf_url = region.urls['pbf']
    PBF_inputfile = download_pbf(geofabrik_pbf_url, update, data_dir)
    country_code = region.short

    # ------- primary file -------
    primary_file_exists = False
    primary_file = os.path.join(data_dir, primary_name, f"{country_code}_{primary_name}.json"
    )
    if os.path.exists(primary_file):
        primary_file_exists = True
        with open(primary_file, encoding="utf-8") as f:
            primary_dict = json.load(f)
    else:
        os.makedirs(os.path.dirname(primary_file), exist_ok=True)

    # TODO: compare update time using metadata in primary_dict
    if not primary_file_exists or update is True:
        primary_dict = run_primary_filter(PBF_inputfile, primary_file, primary_name, mp)

    # ------- feature file -------
    feature_dict = run_feature_filter(primary_dict, feature_name)

    return primary_dict, feature_dict





from collections import namedtuple

Node = namedtuple('Node', ('id', 'tags', 'lonlat'))
"""
A OpenStreetMap `node <https://wiki.openstreetmap.org/wiki/Node>

id: int (node id)
tags: dict (tag names -> tag values)
lonlat: tuple (lon, lat)
"""

Way = namedtuple('Way', ('id', 'tags', 'refs'))
"""
A OpenStreetMap `way <https://wiki.openstreetmap.org/wiki/Way>

id: int (way id)
tags: dict (tag names -> tag values)
refs: list (node ids)
"""

Relation = namedtuple('Relation', ('id', 'tags', 'members'))
"""
A OpenStreetMap `relation <https://wiki.openstreetmap.org/wiki/Relation>

id: int (relation id)
tags: dict (tag names -> tag values)
members: tuple (ref, member_type, role)
    where:
        member_type: str (Node, way, or Relation)
        role: str (role of the member)
"""



__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""CLI interface for earth_osm project.

This module provides a CLI interface for the earth_osm project.

"""

import argparse
import os

from earth_osm.config import primary_feature_element
from earth_osm.eo import get_osm_data
from earth_osm.gfk_data import get_all_valid_list, view_regions


def main():  # pragma: no cover
    """
    The main function executes on commands:
    `python -m earth_osm` and `$ earth_osm `.
    It parses the command line and executes the appropriate function.
    """

    class _HelpAction(argparse._HelpAction):
        def __call__(self, parser, namespace, values, option_string=None):
            parser.print_help()

            # retrieve subparsers from parser
            subparsers_actions = [
                action
                for action in parser._actions
                if isinstance(action, argparse._SubParsersAction)
            ]

            for subparsers_action in subparsers_actions:
                # get all subparsers and print help
                for choice, subparser in subparsers_action.choices.items():
                    print("Subparser '{}'".format(choice))
                    print(subparser.format_help())

            parser.exit()

    # Argparser Notes
    # '+' == 1 or more.
    # '*' == 0 or more.
    # '?' == 0 or 1.

    parser = argparse.ArgumentParser(
        description='Earth-OSM by PyPSA-meets-Earth',
        # epilog='''Example:''',
        add_help=False) # hide default help

    parser.add_argument('-h', '--help', action=_HelpAction,
                        help='earth-osm help')
    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s 0.1.0')

    subparser = parser.add_subparsers(dest='command', required=True, title='Sub Parser', description='''View Supported Regions or Extract OSM Data''')

    extract_parser = subparser.add_parser('extract', help='Extract OSM Data')

    extract_parser.add_argument('primary', choices=primary_feature_element.keys(), type=str, help='Primary Feature')

    # region_group = extract_parser.add_mutually_exclusive_group(required=True) # Get regions using ids or coordinates
    # region_group.add_argument('--regions',nargs="+", type=str, help='Region Identifier')
    # region_group.add_argument('--coords', nargs='+', type=float, help='Region Bounds') #TODO: Change to >3

    extract_parser.add_argument('--regions', nargs="+", type=str, help='Region Identifier') # TODO: replace with region group
    extract_parser.add_argument('--features', nargs="*", type=str, help='Sub-Features')
    extract_parser.add_argument('--update', action='store_true', default=False, help='Update Data')
    extract_parser.add_argument('--mp',  action='store_true', default=True, help='Use Multiprocessing')
    extract_parser.add_argument('--data_dir', nargs="?", type=str, help='Earth Data Directory')
    extract_parser.add_argument('--out_format', nargs="+", type=str, help='Export options')
    extract_parser.add_argument('--out_aggregate', action='store_true', default=False, help='Aggregate Outputs')
    
    view_parser = subparser.add_parser('view', help='View OSM Data')
    view_parser.add_argument('type', help='View Supported', choices=['regions', 'primary'])

    args = parser.parse_args()

    print("\n".join(["",
        "  _   _   _   _   _   _   _   _   _",
        " / \ / \ / \ / \ / \ / \ / \ / \ / \ ",
        "( E | a | r | t | h | - | O | S | M )",
        " \_/ \_/ \_/ \_/ \_/ \_/ \_/ \_/ \_/ "]))

    if args.command == 'view':
        print('Viewing OSM Data')
        view_regions(level=0)
    elif args.command == 'extract':
        if args.regions:
            region_list = list(args.regions)
            if not set(region_list) <= set(get_all_valid_list()):
                raise KeyError(f'Invalid Region {" ".join(set(region_list) - set(get_all_valid_list()))} , run earth_osm view regions to view valid regions')
        # elif args.coords:
        #     # TODO: change coords to shapely polygon, implement geom=True, get regions that way
        #     raise NotImplementedError('Bounding Box Region Identifier Not Implemented')
        if args.out_format:
            out_format = set(args.out_format)
            if not out_format.issubset(['csv', 'geojson']):
                raise KeyError(f'Contains invalid format. Valid formats are: csv, geojson')

        if args.features:
            feature_list = list(args.features)
            if not set(feature_list) <= set(primary_feature_element[args.primary]):
                raise KeyError(f'Invalid Feature {" ".join(set(feature_list) - set(primary_feature_element[args.primary]))}, run earth_osm view features to view valid features')
        else:
            feature_list = primary_feature_element[args.primary]

        if args.data_dir:
            if not os.path.exists(args.data_dir):
                os.makedirs(args.data_dir, exist_ok=True)
            if os.path.isdir(args.data_dir):
                data_dir = args.data_dir
            else:
                raise NotADirectoryError(f'Invalid Data Directory {args.data_dir}')
        else:
            data_dir = os.path.join(os.getcwd(), 'earth_data')

        print('\n'.join(['',
            f'Primary Feature: {args.primary}',
            f'Sub Features: {" - ".join(feature_list)}',
            f'Regions: {" - ".join(region_list)}',
            f'Multiprocessing = {args.mp}',
            f'Update Data = {args.update}',
            f'Data Directory = {data_dir}',
            f'Output Format = {" - ".join(out_format)}',
            f'Aggregate Outputs = {args.out_aggregate}',
            ]))

        get_osm_data(
            region_list=region_list,
            primary_name=args.primary,
            feature_list=feature_list,
            update=args.update,
            mp=args.mp,
            data_dir=data_dir,
            out_format=out_format,
            out_aggregate=args.out_aggregate,
        )

    else:
        # import inquirer #https://github.com/magmax/python-inquirer
        raise NotImplementedError('Interactive Mode Not Implemented')





__author__ = "PyPSA meets Earth"
__copyright__ = "Copyright 2022, The PyPSA meets Earth Initiative"
__license__ = "MIT"

"""
This is the principal module of the earth_osm project.
"""

import logging
import os

import pandas as pd

from earth_osm.config import primary_feature_element, feature_columns
from earth_osm.filter import get_filtered_data
from earth_osm.gfk_data import get_region_tuple
from earth_osm.utils import convert_ways_lines, convert_ways_points, output_creation


logger = logging.getLogger("osm_data_extractor")
logger.setLevel(logging.INFO)

# TODO: Rename to process_region
def process_country(region, primary_name, feature_name, mp, update, data_dir):
    """
    Process Country

    Args:
        region: Region object
        primary_name: Primary Feature Name
        feature_name: Feature Name
        mp: Multiprocessing object
        update: Update flag

    Returns:
        None
    """
    primary_dict, feature_dict = get_filtered_data(region, primary_name, feature_name, mp, update, data_dir)

    primary_data = primary_dict['Data']
    feature_data = feature_dict['Data']

    df_node = pd.json_normalize(feature_data["Node"].values())
    df_way = pd.json_normalize(feature_data["Way"].values())

    element_type = primary_feature_element[primary_name][feature_name]

    if element_type == "way":
        convert_ways_lines(
            df_way, primary_data
        ) if not df_way.empty else logger.warning(
            f"Empty Way Dataframe for {feature_name} in {region.short}"
        )
        if not df_node.empty:
            logger.warning(
                f"Node dataframe not empty for {feature_name} in {region.short}"
            )

    if element_type == "node":
        convert_ways_points(df_way, primary_data) if not df_way.empty else None

    # Add Original Type Column
    df_node["Type"] = "Node"
    df_way["Type"] = "Way"

    # Concatinate Nodes and Ways
    df_feature = pd.concat([df_node, df_way], axis=0)

    # Add Country Column
    # TODO: rename Country to Region
    df_feature["Country"] = region.short

    return df_feature


def get_osm_data(
    region_list=['germany'],
    primary_name='power',
    feature_list=['tower'],
    update=False,
    mp=True,
    data_dir=os.path.join(os.getcwd(), 'earth_data'),
    out_format="csv",
    out_aggregate=False,
):
    """
    Get OSM Data for a list of regions and features
    args:
        region_list: list of regions to get data for
        primary_name: primary feature to get data for
        feature_list: list of features to get data for
        update: update data
        mp: use multiprocessing
    returns:
        dict of dataframes
    """
    region_tuple_list = [get_region_tuple(rs) for rs in region_list]

    for region in region_tuple_list:
        for feature_name in feature_list:
            df_feature = process_country(region, primary_name, feature_name, mp, update, data_dir)
            df_feature = df_feature.reindex(columns=feature_columns[feature_name])
            output_creation(df_feature, primary_name, feature_name, [region], data_dir, out_format, out_aggregate)
